/* -*- mode: c; compile-command: "cd ..; make all"; -*- */
/* flashtool - Programm to flash C167 microcontrollers */

#include <string.h> /* for bzero */
#include <unistd.h> /* for read */
#include <stdlib.h> /* for EXIT_* */
#include <signal.h> /* for signal */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h> /* for struct termios */
#include <stdio.h>

#include "flashtool.h" /* always include this file as first one - it
			  is the global include file */
#include "parse_command_line.h"
#include "error_handling.h"
#include "flashtool_functions.h"

extern const char *command_name;

/* needs to be global for handler function */
int fd; /* file desciptor for serial port */
/* Use this variable to remember original terminal attributes. */
struct termios saved_attributes, saved_attributes_stdin;
int terminal_opened = 0;
int fdstdin; /* file descriptor for stdin */

void allocate_resources (char *device, int baud, int verbose, int noinit);
void free_resources ();
void terminal (int fd, bool verbose); /* serial terminal programm */
void turn_cannonical_on (char *device, int verbose);
void handler (int signal);

int main (int argc, char **argv)
{
  int result = -1;
  struct command_line_options options; /* store the command line
					  switches */
  bool need_serial_line = FALSE; /* TRUE if we need the serial line,
				    determined from options */

  /* generate handlers for signals */
  (void) signal (SIGTERM, handler);
  (void) signal (SIGINT, handler);
  (void) signal (SIGQUIT, handler);
    
  /* Parse the command line and store the switches in struct options */
  options = parse_command_line (argc, argv);
  
  /* check if we have to init the serial line */
  if (options.init || options.terminal || options.info || options.read || options.write || options.bootstrap || options.bootstrap3 || options.flash || options.delete)
    need_serial_line = TRUE;

  /* set the terminal device and store old settings in global
     variable saved_attributes */
  if (need_serial_line)
    allocate_resources (options.device, options.baud, options.verbose, options.noinit);

  /* terminal settings done, now handle input. */
  /* Handle switches and options given on the command line */
  /* And call the executing functioins respectively */
  if (options.hextobin)
    {
       hex_to_bin ( options.verbose, options.input_file_name, options.output_file_name);
    }
  else if (options.convert)
    {
      intel_hex_to_data (options.verbose, options.input_file_name, options.output_file_name);
    }
  else if (options.convert3)
    {
      intel_hex_to_data_bsl3 (options.verbose, options.input_file_name, options.output_file_name);
    }
  else if (options.init)
    {
      initialize_micro_controller (fd, options.verbose);
    }
  else if (options.write)
    {
      int is_ihex = -1;

      initialize_micro_controller (fd, options.verbose);

      /* turn_cannonical_on (options.device, options.verbose); */

      transfer_bootstrap_loader (options.verbose, fd, BSL1, options.output_file_name);
      transfer_bootstrap_loader (options.verbose, fd, BSL2, options.output_file_name);

      is_ihex = is_file_a_intel_hex_file (options.input_file_name);
      if (is_ihex == 0)
	{
	  /* Wenn es ein hex file ist koennen wir es damit
	     transferieren: */
	    transfer_bootstrap_loader_bsl3 (options.verbose, fd, options.input_file_name);   
	    /* _send_flashtool(fd, options.input_file_name, options.verbose);*/
	}
      else 
	{
	  /* wenn es kein intel hex file ist nehmen wir an dass es schon
	     vorher convertiert wurde ( mit switch --convert3) also im
	     binary format vorliegt und wir es direkt transferieren knnen */
	  transfer_bootstrap_loader (options.verbose, fd, options.input_file_name, options.output_file_name);
	}

    }
  else if (options.bootstrap)
    {
      transfer_bootstrap_loader (options.verbose, fd, options.input_file_name, options.output_file_name);
    }
  else if (options.bootstrap3)
    {
      int is_ihex = -1;
      is_ihex = is_file_a_intel_hex_file (options.input_file_name);
      if (is_ihex == 0)
	{
	    /* wenns ein intel hex file is kann es mit dieser funktion
	       transferiert werden */
	  
	  transfer_bootstrap_loader_bsl3 (options.verbose, fd, options.input_file_name); 
	}
      else
	{
	  /* wenn es kein intel hex file ist nehmen wir an dass es schon
	     vorher convertiert wurde ( mit switch --convert3) also im
	     binary format fr bsl3 vorliegt und wir es direkt
	     transferieren knnen */
	  transfer_bootstrap_loader (options.verbose, fd, options.input_file_name, options.output_file_name);
	}
    }
  else if (options.terminal)
    {
      terminal(fd, options.verbose);
    }
  else if (options.flash)
  {
      transfer_ihex_file (options.verbose, fd, options.input_file_name);
  }
  else if (options.delete)
  {
      delete_flash_memory(options.verbose, fd);
  }
  else if (options.read)
  {
      read_flash_memory (options.verbose, fd, options.start, options.stop);
  }
  else 
    {
      print_message(BAIL_OUT, "No option was given, %s --help to print help", command_name);
    }
    
  if (need_serial_line)
    {
      result = tcdrain (fd);
      error_detection (options.verbose, result, NORMAL, "tcdrain %s", options.device);
    }
  
  free_resources ();
  
  exit (EXIT_SUCCESS);
}

void free_resources (void)
{
 /* restore the old port settings */
  if (terminal_opened == 2)
    {
      tcsetattr(fd, TCSAFLUSH, &saved_attributes);
      close(fd);
    }
  else if (terminal_opened == 1)
    {
      close(fd);
    }
  else if (terminal_opened == 3)
    {
      /* Write the old termios structure back */
      tcsetattr(fd, TCSAFLUSH, &saved_attributes);
      tcsetattr (fdstdin, TCSAFLUSH, &saved_attributes_stdin);
      close(fd);
    }  
}

void allocate_resources (char *device, int baud, int verbose, int noinit)
{
  struct termios newtio;                                              
  int result=-1;  
  /*                                                                           
  Open modem device for reading and writing and not as controlling tty       
  because we don't want to get killed if linenoise sends CTRL-C.             
  */                                                                           
  fd = open(device, O_RDWR | O_NOCTTY ); /* set global variable */
  error_detection (verbose, fd, NORMAL, "opening %s", device);
  terminal_opened = 1;

/* Make sure fd is a terminal. */
  result = isatty (fd);
  error_detection (verbose, result, NULLONE, "%s is terminal device", device);

  if (!noinit)
    {
      /* Save the terminal attributes so we can restore them later. */
      result = tcgetattr(fd, &saved_attributes);
      error_detection (verbose, result, NORMAL, "tcgetattr %s", device);
      /* this must be done here, after the attributes have been saved
	 and before they are altered */
      terminal_opened = 2;
     
      /* Set the terminal modes. */
      result = tcgetattr (fd, &newtio);
      error_detection (verbose, result, NORMAL, "tcgetattr %s", device);

      /* Dont clear the struct for new port settings only change the
	 required arguments */
      /* bzero(&newtio, sizeof(newtio)); */
      /*
	Line Speed
	----------
	Set the terminal device to the speed given on the command line
	Better than to set the baudrate direct in the Control Mode of the
	terminal.
      */
#ifndef CYGWIN
      result = cfsetspeed (&newtio, baud);
      error_detection ( verbose, result, NORMAL, "set %s to speed %d",
	 device, baud );
#endif
#ifdef CYGWIN
      result = cfsetospeed (&newtio, BAUDCYGWIN);
      result = cfgetospeed (&newtio);
      if (verbose)
	print_message (MESSAGE, "result: %d", result);
#endif
      /*
	Input Modes
	-----------
	-ICRNL   : DONT map CR to NL (otherwise returned checksum might be wrong
	in cas bsl2 returns 0dH we will get 0aH -> checksum error
      */ 
      newtio.c_iflag &= ~(ICRNL);
      /*
	Output Modes
	------------
	Raw output.
      */
      newtio.c_oflag &= ~(OPOST | OCRNL);
      /*
	Control Modes
	-------------
	BAUDRATE: Set bps rate. We used cfsetspeed to to this (better).
	not set: CRTSCTS : output hardware flow control (only used if the
	cable has all necessary lines. See sect. 7 of Serial-HOWTO)
	CS8     : 8n1 (8bit,no parity,1 stopbit)
	CLOCAL  : local connection, no modem contol 
	CREAD   : enable receiving characters
      */
      newtio.c_cflag |= CS8 | CLOCAL | CREAD;
      /*
	Local Modes
	-----------
	ICANON  : enable canonical input !
	ECHO: disable all echo functionality, and don't send signals to
	calling program
      */
      newtio.c_lflag &= ~(ICANON | ECHO);
      /* newtio.c_lflag = 0; */
/*
	Special Characters
	------------------
	initialize all control characters
	default values can be found in /usr/include/termios.h, and are given       
	in the comments, but we don't need them here
      */
            
      newtio.c_cc[VINTR]    = 0;     /* Ctrl-c  */
      newtio.c_cc[VQUIT]    = 0;     /* Ctrl-\ */
      newtio.c_cc[VERASE]   = 0;     /* del  */
      newtio.c_cc[VKILL]    = 0;     /* @ */
      newtio.c_cc[VEOF]     = 4;     /* Ctrl-d  */
      newtio.c_cc[VTIME]    = 0;     /* inter-character timer unused  */
      newtio.c_cc[VMIN]     = 1;     /* blocking read until 1 character arrives  */
      newtio.c_cc[VSWTC]    = 0;     /* '\0'  */
      newtio.c_cc[VSTART]   = 0;     /* Ctrl-q */
      newtio.c_cc[VSTOP]    = 0;     /* Ctrl-s */
      newtio.c_cc[VSUSP]    = 0;     /* Ctrl-z */
      newtio.c_cc[VEOL]     = 0;     /* '\0' */
      newtio.c_cc[VREPRINT] = 0;     /* Ctrl-r */
      newtio.c_cc[VDISCARD] = 0;     /* Ctrl-u */
      newtio.c_cc[VWERASE]  = 0;     /* Ctrl-w */
      newtio.c_cc[VLNEXT]   = 0;     /* Ctrl-v */
      newtio.c_cc[VEOL2]    = 0;     /* '\0' */
      
      /* Dont clean the line */
      /* now clean the modem line and activate the settings for the port */
      /*result=tcflush(fd, TCIFLUSH);
	error_detection (verbose, result, NORMAL, "flush %s", device);

	result=tcsetattr(fd, TCSANOW, &newtio);
	error_detection (verbose, result, NORMAL, "tcsetattr %s", device);*/

      /* New method without cleaning the terminal */
      result = tcsetattr (fd, TCSAFLUSH, &newtio);
      error_detection (verbose, result, NORMAL, "tcsetattr %s", device);
    }
}

void allocate_stdin_descriptor_for_terminal_program (int verbose)
{
  struct termios term; 
  int result;
  
  /* turn cannonical input processing off for stdin - our terminal */
  fdstdin = STDIN_FILENO;
  result = tcgetattr(fdstdin, &saved_attributes_stdin); /* save
							   terminal
							   attributes */
  error_detection (verbose, result, NORMAL, "tcgetattr %s", "stdin");
  terminal_opened = 3;
  
  result = tcgetattr (fdstdin, &term);
  error_detection (verbose, result, NORMAL, "tcgetattr %s", "stdin");
  term.c_lflag &= ~ICANON;   /* turn cannonical input off */
  result = tcsetattr (fdstdin, TCSAFLUSH, &term);   /* Write the new
						       termios
						       structure back */
  error_detection (verbose, result, NORMAL, "tcsetattr %s", "stdin");
  
}

/* this must should be run after receiving the c167 identification
   byte (in order to recieve this byte the cannonical mode has to be
   turned off!!!) to turn on cannonical input*/
void turn_cannonical_on (char *device, int verbose)
{
  struct termios newtio;                                              
  int result=-1;  

  /* Get the working terminal modes. */
  result = tcgetattr (fd, &newtio);
  error_detection (verbose, result, NORMAL, "tcgetattr %s", device);

  /* alter terminal mode for cannonical input */
  newtio.c_lflag |= ICANON;

  /* Write the new termios structure back */
  result = tcsetattr (fd, TCSAFLUSH, &newtio);
  error_detection (verbose, result, NORMAL, "tcsetattr %s", device);
}

/* handler-function for signal handlers */
void handler (int signal)
{
  free_resources();
  exit (EXIT_SUCCESS);
}


void terminal (int fd, bool verbose) /* cannonical - syncron - wait for mulitple sources with select */
{

#ifndef CYGWIN

  fd_set readfs;    /* file descriptor set */                      
  int    maxfd;     /* maximum file desciptor used */              
  unsigned char terminal, stdin;

  allocate_stdin_descriptor_for_terminal_program(verbose);
  
  maxfd = MAX (fd, fdstdin) + 1;  /* maximum bit entry (fd) to test */  

  printf ("Serial-Terminal-Programm in both directions, C-c to quit\n");
  /* loop until we have a terminating condition */
  while (1) 
    {  
      FD_SET(fd, &readfs);  /* set testing for source 1 */          
      FD_SET(fdstdin, &readfs);  /* set testing for source 2 */          
      /* block until input becomes available */                      
      select(maxfd, &readfs, NULL, NULL, NULL);                      
      if (FD_ISSET(fd, &readfs))         /* input from source 1 available */ 
	{
	    /* Get input from serial terminal and print it on stdout */
	  read (fd, &terminal, 1);   
	  printf("%c", terminal);
	  fflush (stdout);
	}
      if (FD_ISSET(fdstdin, &readfs)) /* input from source 2 available */ 
	{
	    /* get input from stdin and put it on serial terminal line */ 
	  read (fdstdin, &stdin, 1);
	  write (fd, &stdin, 1);
	}
    }

#endif

}

