/* -*- mode: c; compile-command: "cd ..; make all"; -*- */

#include <string.h> /* for bzero */
#include <unistd.h> /* for read */
#include <stdlib.h> /* for EXIT_* */
#include <signal.h> /* for signal */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <stdio.h>

#include "flashtool.h"
#include "error_handling.h"
#include "parse_command_line.h"
#include "flashtool_functions.h"

/* einen character reingeben -> ein integer wert wird zurckgegeben
 eingabe char value = 'c' => ausgabe 13
 eingabe char value = '3' => ausgabe 3 (int)
 es werden also ascii code in werte umgewandelt */
int convert (char value)
{
  switch (value)
	{
	case 0x30 : value = 0x00 ; break ;
	case 0x31 : value = 0x01 ; break ;
	case 0x32 : value = 0x02 ; break ;
	case 0x33 : value = 0x03 ; break ;
	case 0x34 : value = 0x04 ; break ;
	case 0x35 : value = 0x05 ; break ;
	case 0x36 : value = 0x06 ; break ;
	case 0x37 : value = 0x07 ; break ;
	case 0x38 : value = 0x08 ; break ;
	case 0x39 : value = 0x09 ; break ;
	case 0x41 : value = 0x0A ; break ;
	case 0x42 : value = 0x0B ; break ;
	case 0x43 : value = 0x0C ; break ;
	case 0x44 : value = 0x0D ; break ;
	case 0x45 : value = 0x0E ; break ;
	case 0x46 : value = 0x0F ; break ;
	case 0x3A : value = 0x3A ; break ;
	}
  return value ;
}

/* Quik check if the given filename is a file in the intel hex
   format. Returns -1 if not and 0 if it is a intel hex file */
int is_file_a_intel_hex_file (char *filename)
{
  FILE *input_file;
  char collon;

  input_file = fopen (filename, "r");
  if (input_file == NULL)
    print_message (BAIL_OUT, "failed open the file %s for reading", filename);

  collon = fgetc (input_file);
  
  fclose(input_file);

  if (collon != ':') 
    return -1; /* line should beginn with : */
  else
    return 0;
}

/* parses a line of intel hex code, stores the data in bytes[] */
/* and the beginning address in addr, and returns a 0 if the */
/* line was valid, or a - value if an error occured.  The variable */
/* num gets the number of bytes that were stored into bytes[] */
int parse_hex_line(char *theline, int *bytes, int *addr, int *num, int *code, int *cksum)
{
	int sum, len;
	char *ptr;
	
	*num = 0;
	if (theline[0] != ':') return -1; /* line should beginn with : */
	if (strlen(theline) < 11) return -2; /* and should be bigger
					       than 11 characters */
	ptr = theline+1;
	if (!sscanf(ptr, "%02x", &len)) return -3;
	ptr += 2;
	if ( strlen(theline) < (11 + (len * 2)) ) return -4;
	if (!sscanf(ptr, "%04x", addr)) return -5;
	ptr += 4;
	  /* printf("Line: length=%d Addr=%d\n", len, *addr); */
	if (!sscanf(ptr, "%02x", code)) return -6;
	ptr += 2;
	sum = (len & 255) + ((*addr >> 8) & 255) + (*addr & 255) + (*code & 255);
	while(*num != len) {
		if (!sscanf(ptr, "%02x", &bytes[*num])) return -7;
		ptr += 2;
		sum += bytes[*num] & 255;
		(*num)++;
		if (*num >= 256) return -8;
	}
	if (!sscanf(ptr, "%02x", cksum)) return -9;
	if ( ((sum & 255) + (*cksum & 255)) & 255 ) return -10; /* checksum error */
	return 0;
}

void intel_hex_to_data (bool verbose, char *filename, char *output_file_name)
{
  FILE *input_file;
  FILE *output_file;
  char buffer[TRANSFER_LINE_LENGTH];                
  char *result2;
  int result;
  int data[256];
  int address;
  int num, i;
  int code;
  int is_ihex = 0;
  int cksum;

  /* check the input file*/ 
  is_ihex = is_file_a_intel_hex_file (filename);
  if (is_ihex == 0)
    error_detection(verbose, 0, NORESULT, "analyzed that file %s is a intel hex file -> converting to binary", filename);
  else
    print_message (BAIL_OUT, "%s is no intel hex file", filename);

  input_file = fopen (filename, "r");
  if (input_file == NULL)
    print_message (BAIL_OUT, "failed open the file %s for reading", filename);

  output_file = fopen (output_file_name, "w+");
  if (output_file == NULL)
    print_message (BAIL_OUT, "failed open the file %s for reading", "bin.out");

  while (1)
    {
      result2 = fgets (buffer, TRANSFER_LINE_LENGTH, input_file);
      if (result2 == NULL) /* end of file or reading error ? */
	break;
      if (buffer[strlen(buffer)-1] == '\n') buffer[strlen(buffer)-1] = '\0';
      if (buffer[strlen(buffer)-1] == '\r') buffer[strlen(buffer)-1] = '\0';
      result = parse_hex_line(buffer, data, &address, &num, &code, &cksum);
      if (result < 0)
	    print_message (MESSAGE, "error in hex file, error code: %d", result);
      if (code == 0)
      {
	  for (i=0; i<num; i++)
	      fputc ( data[i], output_file);
      }
    }

  if (ferror(input_file)) 
    print_message (BAIL_OUT, "error file %s", filename); 

  if (ferror(output_file)) 
    print_message (BAIL_OUT, "error file %s", filename); 

  fclose(input_file); 
  fclose(output_file); 
}

void intel_hex_to_data_bsl3 (bool verbose, char *filename, char *output_file_name)
{
  FILE *input_file;
  FILE *output_file;
  char buffer[TRANSFER_LINE_LENGTH];                
  char *result2;
  int result;
  int data[256];
  int address;
  int num, i, cksum;
  int code;
  int is_ihex = 0;

  /* check the input file*/
  is_ihex = is_file_a_intel_hex_file (filename);
  if (is_ihex == 0)
    error_detection(verbose, 0, NORESULT, "analyzed that file %s is a intel hex file -> converting to binary", filename);
  else
    print_message (BAIL_OUT, "%s is no intel hex file", filename);
  
  input_file = fopen (filename, "r");
  if (input_file == NULL)
    print_message (BAIL_OUT, "failed open the file %s for reading", filename);

  output_file = fopen (output_file_name, "w+");
  if (output_file == NULL)
    print_message (BAIL_OUT, "failed open the file %s for reading", output_file_name);

  while (1)
    {
      result2 = fgets (buffer, TRANSFER_LINE_LENGTH, input_file);
      if (result2 == NULL) /* end of file or reading error ? */
	break;
      if (buffer[strlen(buffer)-1] == '\n') buffer[strlen(buffer)-1] = '\0';
      if (buffer[strlen(buffer)-1] == '\r') buffer[strlen(buffer)-1] = '\0';
      
      result = parse_hex_line(buffer, data, &address, &num, &code, &cksum);
      if (result < 0)
	    print_message (MESSAGE, "error in hex file, error code: %d", result);
       
      if (code == 0 )
      {
	fputc (code, output_file); //gib code aus 1 byte
	fputc (num, output_file); //lanege ausgeben 1 byte

	if (address <= 255)
	    fputc (0, output_file);
	fputc (address, output_file); //addrese ausgeben 2 byte
      
	for (i=0; i<num; i++)
	    fputc ( data[i], output_file);
      }
      else if (code == 1)
	  fputc (code, output_file); //gib code aus 1 byte

      else if (code == 2)
      {
	  fputc(code, output_file);
	  fputc(data[0], output_file);
	  //fputc(4, output_file);
	  //fputc(data[1], output_file);
      }
      //fputc('\n', output_file);
    }

  if (ferror(input_file)) 
    print_message (BAIL_OUT, "error file %s", filename); 

  if (ferror(output_file)) 
    print_message (BAIL_OUT, "error file %s", filename); 

  fclose(input_file); 
  fclose(output_file); 
}


void hex_to_bin (bool verbose, char *filename, char *output_file_name)
{
  FILE *input_file;
  FILE *output_file;
  char char_high, char_low;
  int bin_high, bin_low, bin;
  int is_ihex = 0;

  /* check the input file*/
  is_ihex = is_file_a_intel_hex_file (filename);
  if (is_ihex == -1)
    error_detection(verbose, 0, NORESULT, "analyzed that file %s is a raw hex file -> converting", filename);
  else
    print_message (BAIL_OUT, "%s is a intel hex file but this conversion only works with raw hex files", filename);


  /* IO with streams  */
  input_file = fopen (filename, "r");
  if (input_file == NULL)
    print_message (BAIL_OUT, "failed open the file %s for reading", filename);

  output_file = fopen (output_file_name, "w+");
  if (output_file == NULL)
    print_message (BAIL_OUT, "failed open the file %s for reading", output_file_name);

  while (1)
    {
        char_high=fgetc(input_file);
	char_low=fgetc(input_file);
        if (char_high==EOF || char_low==EOF)
	    break;
	if (char_high == '\n')
	  {
	    char_high = char_low;
	    char_low = fgetc(input_file);
	  }
	else if (char_low == '\n')
	  {
	    char_low = fgetc(input_file);
	  }
	  
	bin_high = convert (char_high);
	bin_low = convert (char_low);
	bin = (bin_high << 4) | bin_low; /* conversion */
	fputc (bin, output_file);
    }
  
  if (ferror(input_file)) 
    print_message (BAIL_OUT, "error file %s", filename); 

  if (ferror(output_file)) 
    print_message (BAIL_OUT, "error file %s", filename); 
  
  fclose(input_file); 
  fclose(output_file); 

}

void initialize_micro_controller (int fd, bool verbose)
{
  ssize_t result;
  unsigned char answer;

 /* initialize C167 - send ASCII 0 */
 result = write (fd, "\0", 1);
 error_detection (verbose, result, NORMAL, "sending ASCII 0 (uC should now return C5H = 197d)");
 
 /* read answer */
 result = read (fd, &answer, 1);
 error_detection (verbose, result, NORMAL, "reading answer");
 if (verbose==1)
   {  
     if (answer == 0xC5 ) 
       print_message (MESSAGE, "microcontroller answered C5H -> C167 present");
     else if (answer == 0xA5 ) 
       print_message (MESSAGE, "microcontroller answered A5H -> previous version of C167");
     else if (answer == 0x55 ) 
       print_message (MESSAGE, "microcontroller answered 55H  -> 8xC166 present");
     else if (answer == 0xB5 ) 
       print_message (MESSAGE, "microcontroller answered B5H  -> C161 or C165 present");
     else 
       print_message (WARNING, "microcontroller answered %02xH - unknown microcontroller", answer);
   } 
}


void transfer_bootstrap_loader (bool verbose, int fd, char *filename, char *output_file_name)
{
  size_t result;
  int sign;                
  FILE *input_file;
  int checksum = 0;
  int is_ihex = 0;

  /* check the input file*/
  is_ihex = is_file_a_intel_hex_file (filename);
  if (is_ihex == -1)
    error_detection(verbose, 0, NORESULT, "analyzed that file %s is a binary file -> flashing", filename);
  else
    {
      error_detection(verbose, 0, NORESULT, "%s is a intel hex file -> converting to binary -> flashing", filename);
      
      intel_hex_to_data (verbose, filename, output_file_name);
      strncpy (filename, output_file_name, FILENAME_MAX_LENGTH);
    }

  input_file = fopen (filename, "r");
  if (input_file == NULL)
    print_message (BAIL_OUT, "failed open the file %s for reading", filename);

  while (1)
    {
      sign=fgetc(input_file);
      if (sign==EOF)
	{
	  if (verbose) //terminate line
	    printf("\n");
	  break;
	}

      result = write (fd, &sign ,1);
      if (verbose)
      {
	  printf("%02x", sign);
      }
      if (result < 0)
	print_message (BAIL_OUT, "cannot write");
    }
  
  if (ferror(input_file)) 
    print_message (BAIL_OUT, "error file %s", filename); 
  fclose(input_file); 

  read (fd, &checksum, 1);
  if (verbose)
    {
      print_message (MESSAGE, "got %02xH from microcontroller", checksum);
    }
}

void transfer_bootstrap_loader_bsl3 (bool verbose, int fd, char *filename)
{
  ssize_t result;
  char buffer[TRANSFER_LINE_LENGTH];                
  int i , num;
  FILE *input_file;
  char *result2;
  int data[256];
  int address;
  int code, cksum;
  int is_ihex = 0;
  unsigned char answer = 0;
  int low;
  int high;
  
  /* check the input file*/
  is_ihex = is_file_a_intel_hex_file (filename);
  if (is_ihex == 0)
    error_detection(verbose, 0, NORESULT, "analyzed that file %s is a intel hex file -> converting to binary and sending directly", filename);
  else
    print_message (BAIL_OUT, "%s is no intel hex file", filename);
  
  input_file = fopen (filename, "r");
  if (input_file == NULL)
    print_message (BAIL_OUT, "failed open the file %s for reading", filename);

  while (1)
    {
	result2 = fgets (buffer, TRANSFER_LINE_LENGTH, input_file); // Zeile von Hexfile einlesen
	if (result2 == NULL) /* end of file or reading error ? */
	    break;

	// In Binaer umwandeln
	if (buffer[strlen(buffer)-1] == '\n') buffer[strlen(buffer)-1] = '\0';
	if (buffer[strlen(buffer)-1] == '\r') buffer[strlen(buffer)-1] = '\0';
	result = parse_hex_line(buffer, data, &address, &num, &code, &cksum);
	if (result < 0)
	    print_message (WARNING, "error in hex file, error code: %d", result);
       
	if (code == 1)
	  {
	    write (fd, &code, 1); //gib code aus 1 byte
	    if (verbose) printf("%02x", code);
	  }
	else if (code == 2 || code == 04)
	  {
	    write(fd, &code, 1);
	    if (verbose) printf("%02x", code);
	    write (fd, &data[0],1 );
	    if (verbose) printf("%02x", data[0]);
	    //write (fd, (int) 4, 1); //wahrscheinlich nich mehr noetig
	    //write (fd, &data[1], 1);
	  }
	else // code == 0 
	  {
	    write (fd, &code, 1); //gib code aus 1 byte
	    if (verbose) printf("%02x", code);
	    
	    write (fd, &num, 1); //lanege ausgeben 1 byte
	    if (verbose) printf("%02x", num);
	    
	    //Addresse in high und low ausmaskieren und ausgeben!!!
	    high = address & 0xFF00;
	    high = high >> 8;
	    low = address & 0xFF;
	    write (fd, &high, 1); //addrese ausgeben 1 byte
	    if (verbose) printf("%02x", high);
	    write (fd, &low, 1); //addrese ausgeben 1 byte
	    if (verbose) printf("%02x", low);
	    //  write (fd, &address, 2); //Achtung mit dieser Zeile
	    //  statt dem code drber gehts nicht !!!!!!!!!!!!!
	    
	    for (i=0; i<num; i++)
	      {
		write (fd, &data[i], 1);
		if (verbose) printf("%02x", data[i]);
	      }
	  }
	
	//checksum lesen und mit richtiger vergleichen
	read (fd, &answer, 1);
	if (verbose)
	  printf (" checksum %02x", answer); 
	/* bei code == 2 und 4 stimmt checksum nicht */
	if ( (answer != cksum) && !((code ==2) || (code == 4)) ) 
	  {
	    if (verbose)
	      printf (" CHECKSUM ERROR\n");
	    print_message (WARNING, "checksum error");
	  }
	else
	  { 
	    if (verbose)
	      printf (" OK\n");
	  }
   }
  
  if (ferror(input_file)) 
    print_message (BAIL_OUT, "error file %s", filename); 
  fclose(input_file); 
}

void send_command (bool verbose, int fd, unsigned char *command)
{
  int result;
  unsigned char receive;

  result = write (fd, command, strlen(command));
  error_detection (verbose, result, NORMAL, "write command %s", command);
  
  /* read answers */
  /* checksum */
  result = read (fd, &receive, 1);
  if (receive == 'p') 
    {
      if (verbose)
	print_message (MESSAGE, "received %c, checksum ok", receive);
    }
  else if (receive == 'P') 
    print_message (WARNING, "received %c, checksum error", receive);
  else 
    print_message (WARNING, "received %c, uC programm error", receive);
	 
  /* read second answer */
  read  (fd, &receive, 1); 
  if (receive == 'e') 
    {
      if (verbose)
	print_message (MESSAGE, "received %c, extended address ok", receive);
    }
  else if (receive == 'x')
    {
      if (verbose)
	print_message (MESSAGE, "received %c, delete eprom ok", receive);
    }
  else if (receive == 's') 
    {
      if (verbose)
	print_message (MESSAGE, "received %c, end of file ok", receive);
    }
  else if (receive == 'w')
    {
      if (verbose)
	print_message (MESSAGE, "received %c, write to eprom ok", receive);
    }
  else if (receive == 't') 
    print_message (WARNING, "received %c, error tranfering file", receive);
  else
    print_message (WARNING, "received %c, uC programm error", receive);

  if (receive != 't')
    {
      /* read third answer */
      read  (fd, &receive, 1);
      if (receive == 't') 
	{
	  if (verbose)
	    print_message (MESSAGE, "received %c, ok for next line", receive);
	}
      else 
	print_message (WARNING, "received %, uC programm error", receive);
    }

  fflush(stdout);
}

void delete_flash_memory(bool verbose, int fd)
{
    unsigned char delcommand[14] = ":00000004FC\n";
    send_command (verbose, fd, delcommand);
}

void read_flash_memory(bool verbose, int fd, int start, int stop)
{
  char buffer[20];
  int sum;
  unsigned char rec;

  sum = 2 + ((start>>8)&255) + (start&255) + 3 + ((stop>>8)&255) + (stop&255);
  sprintf (buffer, ":02%04X03%04X%02X\n", start, stop, (-sum)&255);
  
  //printf ("schreibe %s", buffer);
  write (fd, &buffer, strlen(buffer));
  
  do {
    read(fd, &rec, 1);
    printf("%c", rec);
  } while (rec!='r'); 

}


void transfer_ihex_file (bool verbose, int fd, char *filename)
{
  FILE *input_file;
  char line [TRANSFER_LINE_LENGTH]; 
  //unsigned char rec;

  input_file = fopen (filename, "r");
  if (input_file == NULL)
    print_message (BAIL_OUT, "failed open the file %s for reading", filename);

  while (fgets (line, TRANSFER_LINE_LENGTH, input_file))
    {
      send_command (verbose, fd, line);
      /* write (fd, line, strlen(line));
      sleep(1);
      do {
        read(fd, &rec, 1);
        printf("%c", rec);
	} while (rec!='\r'); */
      
    } 

  if (ferror(input_file)) 
    print_message (BAIL_OUT, "error file %s", filename); 
  fclose(input_file); 
}

