/* -*- mode: c; compile-command: "cd ..; make all"; -*- */

#include <getopt.h> /* For getopt_long */
#include <stdio.h>  /* For *printf */
#include <string.h> /* For strncpy */
#include <stdlib.h> /* For EXIT_SUCCESS */

#include "flashtool.h" /* always include this file as first one - it is the global include file */
#include "error_handling.h"
#include "parse_command_line.h"

const char *command_name = "<not yet set>";  /* command name */

struct command_line_options parse_command_line (int argc, char **argv)
{
  int c;
  int ret=-1;
  bool output_file_is_given = FALSE;
  bool input_file_is_given = FALSE;

  /* initialization of the command line arguments - set it to the
     preset values (macros in header file flashtool.h) */
  static struct command_line_options options = { BAUD,
						 DEVICE,
						 0,
						 0,
						 FALSE,
						 FALSE,
						 FALSE,
						 FALSE,
						 FALSE,
						 FALSE,
						 FALSE,
						 FALSE,
						 FALSE, 
						 FALSE,
						 FALSE,
						 FALSE,
						 FALSE,
						 FALSE,
						 FALSE,
						 FALSE,
						 OUTPUT,
						 INPUT};

  /* initialisation of baudrate and device file */
  options.baud = BAUD;
  strncpy (options.device, (const char *) DEVICE, (size_t) FILENAME_MAX_LENGTH);

  /* save the command name in a global variable */
  command_name = argv[0];

  /* bail out if there are no arguments given on the command line */
  if (argc <= 1)
    {
      print_message (BAIL_OUT, "No command line arguments, %s --help to print help", command_name);
    }

  while (1)
    {
      static struct option long_options[] =
        {
          /* These options set a flag. */
          {"verbose", no_argument, &options.verbose, 1},
          {"version", no_argument, &options.version, 1},
          {"terminal", no_argument, &options.terminal, 1},
          {"help",   no_argument, &options.help, 1},
          {"write",  no_argument, &options.write, 1},
          {"flash",  no_argument, &options.flash, 1},
          {"delete",  no_argument, &options.delete, 1},
          {"bootstrap",  no_argument, &options.bootstrap, 1},
	  {"bootstrap3",  no_argument, &options.bootstrap3, 1},
	  {"init",  no_argument, &options.init, 1},
          {"hextobin",  no_argument, &options.hextobin, 1},
	  {"convert",  no_argument, &options.convert, 1},
	  {"convert3",  no_argument, &options.convert3, 1},
	  {"noinit", no_argument, &options.noinit, 1},
	  {"read",  no_argument, &options.read, 1},
          {"info",  no_argument, &options.info, 1},
          /* These options don't set a flag.
             We distinguish them by their indices. */
          {"start",    required_argument, 0, 's'},
          {"stop",    required_argument, 0, 't'},
          {"output",    required_argument, 0, 'o'},
          {"baud",    required_argument, 0, 'b'},
          {"device",    required_argument, 0, 'd'},
          {0, 0, 0, 0}
        };
      /* `getopt_long' stores the option index here. */
      int option_index = 0;

      c = getopt_long (argc, argv, "s:t:o:b:d:", long_options, &option_index);
      /* Detect the end of the options. */
      if (c == -1)
        break;

      switch (c)
        {
        case 0:
          /* If this option set a flag, do nothing else now. */
          if (long_options[option_index].flag != 0)
            break;
          error_detection (options.verbose, 0, NORESULT, "getopt-long: option %s", long_options[option_index].name);
          if (optarg)
            error_detection (options.verbose, 0, NORESULT, " with arg %s", optarg);
          break;
        case 's':
	  error_detection (options.verbose, 0, NORESULT, "option -s with value `%s'", optarg);
	  ret=sscanf (optarg, "%d", &options.start);
	  if (ret<1)
	    error_detection (0, 0, NORESULT_BAILOUT, "You must use a valid start address eg. 0");
          break;
        case 't':
	  error_detection (options.verbose, 0, NORESULT, "option -t with value `%s'", optarg);
	  ret=sscanf (optarg, "%d", &options.stop);
	  if (ret<1)
	    error_detection (0, 0, NORESULT_BAILOUT, "You must use a valid stop address eg. 22");

          break;
        case 'o':
	  error_detection (options.verbose, 0, NORESULT, "option -o with value `%s'\n", optarg);
	  strncpy (options.output_file_name, optarg, FILENAME_MAX_LENGTH);
	  output_file_is_given = TRUE;
          break;
        case 'b':
	  if (optarg)
	    {
	      error_detection (options.verbose, 0, NORESULT, "option -b with value `%s'", optarg);
	      ret=sscanf (optarg, "%d", &options.baud);
	      if (ret<1)
		error_detection (0, 0, NORESULT_BAILOUT, "You must use a valid baudrate eg. 38400");
	    }
	  break;
        case 'd':
	  error_detection (options.verbose, ret, NORESULT, "option -d with value `%s'", optarg);
	  strncpy (options.device, optarg, FILENAME_MAX_LENGTH);
          break;
        case '?':
          /* `getopt_long' already printed an error message. */
          break;
        default:
          abort ();
        }
    }
  if (options.version)
    {
      print_version ();
      exit (EXIT_SUCCESS);
    }
  if (options.help)
    {
      print_help ();
      exit (EXIT_SUCCESS);
    }
  
  /* Print any remaining command line arguments (not options). */
  if (optind < argc)
    {
      if ((optind+1) < argc)
	{
	  print_help ();
	  print_message (BAIL_OUT, "Too many command line arguments");
	}
      else
	{
	  input_file_is_given = TRUE;
	  strncpy ( options.input_file_name, argv[optind], FILENAME_MAX_LENGTH);
	  if (options.verbose)
	    print_message ( MESSAGE, "Input File: %s", options.input_file_name);
	}
    }

  /* Initialize intelligent output and input file name depending on
  options */
  if (options.write && (input_file_is_given == FALSE))
    strncpy ( options.input_file_name, BSL3, FILENAME_MAX_LENGTH);
  if ((options.convert || options.convert3 || options.write || options.bootstrap) && (output_file_is_given == FALSE))
    {
      const char delimiters[] = ".";
      char *token, *cp;
      
      cp = strdup (options.input_file_name);
      token = strtok (cp, delimiters);
      strncpy (options.output_file_name, token, FILENAME_MAX_LENGTH);
      if (options.convert || options.bootstrap)
	strncat (options.output_file_name, ".bin", FILENAME_MAX_LENGTH);
      if (options.convert3 || options.write)
	strncat (options.output_file_name, ".bin3", FILENAME_MAX_LENGTH);
    }
     

  return options;
}

